import asyncio
import dill
import psutil
import signal
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from hedra.core.engines.client.config import Config
from hedra.core.engines.client.time_parser import TimeParser
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.playwright import (
    MercuryPlaywrightClient,
    ContextConfig
)
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.optimize.optimization.algorithms import registered_algorithms
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.condition.decorator import condition
from hedra.core.hooks.types.context.decorator import context
from hedra.core.hooks.types.event.decorator import event
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.graphs.stages.optimize.optimization import Optimizer, DistributionFitOptimizer
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.data.serializers import Serializer
from hedra.logging import logging_manager
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.monitoring.base.exceptions import MonitorKilledError
from hedra.plugins.extensions import get_enabled_extensions
from hedra.plugins.types.extension.types import ExtensionType
from hedra.plugins.types.extension.extension_plugin import ExtensionPlugin
from hedra.plugins.types.plugin_types import PluginType
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes
from typing import (
    Dict, 
    List, 
    Tuple, 
    Any, 
    Union,
    Optional,
    Type
)
from .optimization.parameters import Parameter
from .parallel import optimize_stage


BatchedOptimzationCandidates = List[Tuple[str, Execute, int]] 

OptimizeParameterPair = Tuple[Union[int, float], Union[int, float]]

MonitorResults = Dict[str, List[Union[int, float]]]


def handle_loop_stop(
    signame,
    executor: ThreadPoolExecutor
):
    try:
        executor.shutdown(cancel_futures=True)

    except KeyboardInterrupt:
        raise RuntimeError()

    except BrokenPipeError:
        raise RuntimeError()
        
    except RuntimeError:
        raise RuntimeError()
    
    except MonitorKilledError:
        raise RuntimeError()



class Optimize(Stage):
    stage_type=StageTypes.OPTIMIZE
    optimize_iterations=0
    algorithm='shg'
    time_limit='1m'
    optimize_params: List[Parameter]=[
        Parameter(
            'batch_size',
            minimum=0.5,
            maximum=2,
            feed_forward=True
        )
    ]
    priority: Optional[str]=None
    retries: int=0
    
    def __init__(self) -> None:
        super().__init__()
        self.generation_optimization_candidates = 0
        self.execution_stage_id = 0

        self.results = None

        time_parser = TimeParser(self.time_limit)
        self.stage_time_limit = time_parser.time
        self.requires_shutdown = True
        self.allow_parallel = True

        self.optimization_execution_time_start = 0
        self.optimization_execution_time = 0
        self.accepted_hook_types = [ 
            HookType.CONTEXT,
            HookType.EVENT, 
            HookType.TRANSFORM 
        ]

        self.optimize_iterations = self.optimize_iterations
        self.algorithm = self.algorithm
        self.time_limit = self.time_limit
        self.optimize_params = self.optimize_params

        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._thread_executor: Union[ThreadPoolExecutor, None] = None

        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.stage_retries = self.retries
        self.serializer = Serializer()

    @Internal()
    async def run(self):

        await self.setup_events()
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)

    @context()
    async def collect_optimization_stages(
        self,
        setup_stage_configs: Dict[str, Config] = {},
        optimize_stage_candidates: Dict[str, Execute]={},
        setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]]={},
        execute_stage_streamed_analytics: Dict[str, List[StreamAnalytics]]={}

    ):
        main_monitor_name = f'{self.name}.main'

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        cpu_monitor.stage_type = StageTypes.OPTIMIZE
        memory_monitor.stage_type = StageTypes.OPTIMIZE

        await cpu_monitor.start_background_monitor(main_monitor_name)
        await memory_monitor.start_background_monitor(main_monitor_name)

        self.optimization_execution_time_start = time.monotonic()

        self.context.ignore_serialization_filters = [
            'optimize_stage_workers_map',
            'optimize_stage_batched_stages',
            'optimize_stage_candidates',
            'optimize_stage_monitors',
            'setup_stage_ready_stages',
            'execute_stage_setup_hooks',
            'execute_stage_results',
            'session_stage_monitors'
        ]

        stage_names = ', '.join(list(optimize_stage_candidates.keys()))

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Optimizing stages {stage_names} using {self.algorithm} algorithm')
        await self.logger.spinner.append_message(f'Optimizer - {self.name} optimizing stages {stage_names} using {self.algorithm} algorithm')

        optimize_stages = [(
            stage.name, 
            stage
        ) for stage in optimize_stage_candidates.values()]


        stages_count = len(optimize_stage_candidates)

        # We may have less workers available during the optimize stage than assigned
        # to the execute stage, so store the original workers count for later.
        stage_workers_map = {
            stage.name: stage.workers for stage in optimize_stage_candidates.values()
        }

        batched_stages: BatchedOptimzationCandidates = list(self.executor.partion_stage_batches(optimize_stages))

        return {
            'setup_stage_experiment_config': setup_stage_experiment_config,
            'setup_stage_configs': setup_stage_configs,
            'execute_stage_streamed_analytics': execute_stage_streamed_analytics,
            'optimize_stage_candidates': optimize_stage_candidates,
            'optimize_stage_stage_names': stage_names,
            'optimize_stage_stages_count': stages_count,
            'optimize_stage_workers_map': stage_workers_map,
            'optimize_stage_batched_stages': batched_stages,
            'optimize_stage_monitors': {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            }
        }
    
    @event()
    async def get_stage_plugins(self):

        optimize_plugins = self.plugins_by_type.get(PluginType.OPTIMIZER)
        for plugin_name, plugin in optimize_plugins.items():
            registered_algorithms[plugin_name] = lambda plugin_config: plugin(plugin_config)
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Optimizer plugin - {plugin_name}')
  
        execute_stage_extensions: Dict[str, ExtensionPlugin] = {}

        stage_extension_plugins: List[str] = self.plugins_by_type.get(
            PluginType.EXTENSION, {}
        )

        extension_plugins: Dict[str, Type[ExtensionPlugin]] = get_enabled_extensions()
        for plugin_name in stage_extension_plugins:
            plugin = extension_plugins.get(plugin_name)

            if plugin:
                enabled_plugin = plugin()
                execute_stage_extensions[enabled_plugin.name] = enabled_plugin

        source_stage_plugins = defaultdict(list)
        for plugin in self.plugins.values():
            source_stage_plugins[plugin.type].append(plugin.name)

        return {
            'optimize_stage_plugins': source_stage_plugins,
            'optimize_stage_extensions': execute_stage_extensions
        }
    
    @event()
    async def collect_loaded_actions(
        self,
        optimize_stage_candidates: Dict[str, Execute]={},
    ):
        loaded_actions: Dict[str, List[ActionHook]] = defaultdict(list)

        for candidate_stage in optimize_stage_candidates.values():

            for value in candidate_stage.context.values():

                if isinstance(value, ActionHook):
                    loaded_actions[candidate_stage.name].append(value)

                elif isinstance(value, list):

                    for item in value:
                        if isinstance(item, ActionHook):
                            loaded_actions[candidate_stage.name].append(item)

                elif isinstance(value, dict):

                    for item in value.values():
                        if isinstance(item, ActionHook):
                            loaded_actions[candidate_stage.name].append(item)

        return {
            'optimize_stage_loaded_actions': loaded_actions
        }

    
    @condition(
        'collect_optimization_stages',
        'get_stage_plugins'
    )
    async def check_has_multiple_workers(self):
        return {
            'optimize_stage_has_multiple_workers': self.total_pool_cpus > 1
        }


    @event('check_has_multiple_workers')
    async def create_optimization_configs(
        self,
        setup_stage_configs: Dict[str, Config] = {},
        optimize_stage_stages_count: int=0,
        optimize_stage_batched_stages: BatchedOptimzationCandidates=[],
        setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]]={},
        execute_stage_streamed_analytics: Dict[str, List[StreamAnalytics]]={},
        optimize_stage_loaded_actions: Dict[str, List[ActionHook]]=[],
        optimize_stage_has_multiple_workers: bool = False
    ):
        if optimize_stage_has_multiple_workers:

            batched_configs = []

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Batching optimization for - {optimize_stage_stages_count} stages')
            
            serializable_context = self.context.as_serializable()

            for stage_name, stage, assigned_workers_count in optimize_stage_batched_stages:

                configs = []
                selected_stage_config = setup_stage_configs.get(stage.name) 

                batch_size = int(selected_stage_config.batch_size/assigned_workers_count)

                stage_loaded_actions = optimize_stage_loaded_actions.get(stage_name)
                loaded_actions: List[str] = [
                    self.serializer.serialize_action(
                        action_hook
                    ) for action_hook in stage_loaded_actions
                ]

                for worker_id in range(assigned_workers_count):

                    execute_stage_plugins = defaultdict(list)

                    for plugin in stage.plugins.values():
                        execute_stage_plugins[plugin.type].append(plugin.name)


                    configs.append({
                        'graph_name': self.graph_name,
                        'graph_path': self.graph_path,
                        'graph_id': self.graph_id,
                        'enable_unstable_features': active_flags[FlagTypes.UNSTABLE_FEATURE],
                        'logfiles_directory': logging_manager.logfiles_directory,
                        'log_level': logging_manager.log_level_name,
                        'worker_id': worker_id,
                        'source_stage_context': {
                            context_key: context_value for context_key, context_value in serializable_context
                        },
                        'source_stage_name': self.name,
                        'source_stage_id': self.stage_id,
                        'execute_stage_name': stage_name,
                        'setup_stage_experiment_config': setup_stage_experiment_config,
                        'execute_stage_loaded_actions': loaded_actions,
                        'execute_stage_streamed_analytics': execute_stage_streamed_analytics,
                        'execute_stage_generation_count': assigned_workers_count,
                        'execute_stage_id': stage.execution_stage_id,
                        'execute_stage_config': selected_stage_config,
                        'execute_stage_batch_size': batch_size,
                        'execute_setup_stage_name': stage.context['execute_stage_setup_by'],
                        'execute_stage_plugins': execute_stage_plugins,
                        'optimizer_params': self.optimize_params,
                        'optimizer_iterations': self.optimize_iterations,
                        'optimizer_algorithm': self.algorithm,
                        'optimize_stage_workers': self.workers,
                        'time_limit': self.stage_time_limit
                    })

                configs[assigned_workers_count-1]['execute_stage_batch_size'] += batch_size%assigned_workers_count

                configs = [
                    dill.dumps(config) for config in configs
                ]

                batched_configs.append((
                    stage_name,
                    assigned_workers_count,
                    configs
                ))

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Provisioned - {assigned_workers_count} - workers for stage - {stage_name}')

            return {
                'optimize_stage_batched_configs': batched_configs
            }

    @event('create_optimization_configs')
    async def execute_batched_optimization(
        self,
        optimize_stage_stages_count: int=0,
        optimize_stage_batched_configs: Dict[str, Any]={},
        optimize_stage_has_multiple_workers: bool = False
        
    ):
        if optimize_stage_has_multiple_workers:

            optimization_results = []
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting optimizaiton for - {optimize_stage_stages_count} - stages')

            results = await self.executor.execute_batches(
                optimize_stage_batched_configs,
                optimize_stage
            )


            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed optimizaiton for - {optimize_stage_stages_count} - stages')

            for _, result_batch in results:
                optimization_results.extend([
                    dill.loads(results_set) for results_set in result_batch
                ])

            return {
                'optimize_stage_results': optimization_results
            }
        
    @event('check_has_multiple_workers')
    async def setup_single_worker_optimization(
        self,
        setup_stage_configs: Dict[str, Config]={},
        optimize_stage_extensions: Dict[str, ExtensionPlugin] = {},
        optimize_stage_has_multiple_workers: bool = False,
        optimize_stage_candidates: Dict[str, Execute]={},
        
    ):
        if optimize_stage_has_multiple_workers is False:

            self._loop = asyncio.get_event_loop()
            self._thread_executor = ThreadPoolExecutor(
                max_workers=psutil.cpu_count(logical=False)
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):

                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._thread_executor
                    )
                )

            for stage_name, persona_config in setup_stage_configs.items():

                execute_stage: Execute = optimize_stage_candidates.get(stage_name)

                for extension in optimize_stage_extensions.values():

                    if extension.extension_type == ExtensionType.GENERATOR:
                        results = await extension.execute(**{
                            'execute_stage_name': execute_stage.name,
                            'execute_stage_hooks': execute_stage.hooks,
                            'persona_config': persona_config
                        })

                        execute_stage = results.get('execute_stage')

                        if execute_stage:
                            execute_stage.hooks = results.get('execute_stage_hooks')

                action_and_task_hooks: List[Union[ActionHook, TaskHook]] = [
                    *execute_stage.hooks[HookType.ACTION],
                    *execute_stage.hooks[HookType.TASK]
                ]

                for hook in action_and_task_hooks:
                    if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

                        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Setting up Playwright Session')

                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
                        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')


                        await hook.session.setup(
                            config=ContextConfig(
                                browser_type=persona_config.browser_type,
                                device_type=persona_config.device_type,
                                locale=persona_config.locale,
                                geolocation=persona_config.geolocation,
                                permissions=persona_config.permissions,
                                color_scheme=persona_config.color_scheme,
                                options=persona_config.playwright_options
                            )
                        )

                optimize_stage_candidates[stage_name] = execute_stage

            return {
                'optimize_stage_candidates': optimize_stage_candidates
            }

    @event('setup_single_worker_optimization')
    async def execute_single_worker_optimization(
        self,
        setup_stage_configs: Dict[str, Config]={},
        optimize_stage_candidates: Dict[str, Execute]={},
        optimize_stage_has_multiple_workers: bool = False,
        execute_stage_streamed_analytics: Dict[str, List[StreamAnalytics]]={}

    ):
        
        optimizers: List[Union[DistributionFitOptimizer, Optimizer]] = []

        if optimize_stage_has_multiple_workers is False:
            for stage_name, persona_config in setup_stage_configs.items():

                execute_stage: Execute = optimize_stage_candidates.get(stage_name)

                if persona_config.experiment and persona_config.experiment.get('distribution'):

                    optimizer = DistributionFitOptimizer({
                        'graph_name': self.graph_name,
                        'graph_id': self.graph_id,
                        'source_stage_name': self.name,
                        'source_stage_id': self.stage_id,
                        'params': self.optimize_params,
                        'stage_name': stage_name,
                        'stage_config': persona_config,
                        'stage_hooks': execute_stage.hooks,
                        'iterations': self.optimize_iterations,
                        'algorithm': self.algorithm,
                        'time_limit': self.time_limit,
                        'stream_analytics': execute_stage_streamed_analytics
                    })

                else:
                    optimizer = Optimizer({
                        'graph_name': self.graph_name,
                        'graph_id': self.graph_id,
                        'source_stage_name': self.name,
                        'source_stage_id': self.stage_id,
                        'params': self.optimize_params,
                        'stage_name': stage_name,
                        'stage_config': persona_config,
                        'stage_hooks': execute_stage.hooks,
                        'iterations': self.optimize_iterations,
                        'algorithm': self.algorithm,
                        'time_limit': self.time_limit,
                        'stream_analytics': execute_stage_streamed_analytics
                    })

                optimizers.append(optimizer)

            results: List[Dict[str, Union[int, str, float]]] = await asyncio.gather(*[
                self._loop.run_in_executor(
                    self._thread_executor,
                    optimizer.optimize
                ) for optimizer in optimizers
            ])

            self._thread_executor.shutdown(cancel_futures=True)

            optimization_results: List[Dict[str, Union[int, str, float]]] = []
            for optimization_result in results:

                target_stage_name = optimization_result.get('optimize_target_stage')
                stage_config = setup_stage_configs.get(target_stage_name)
                

                if stage_config.experiment and optimization_result.get('optimized_distribution'):
                    stage_config.experiment['distribution'] = optimization_result.get('optimized_distribution')

                stage_config.batch_size = optimization_result.get('optimized_batch_size', stage_config.batch_size)
                stage_config.batch_interval = optimization_result.get('optimized_batch_interval', stage_config.batch_interval)
                stage_config.batch_gradient = optimization_result.get('optimized_batch_gradient', stage_config.batch_gradient)

                await self.logger.filesystem.aio['hedra.optimize'].info(f'{optimizer.metadata_string} - Optimization complete')

                
                optimization_results.append({
                    'stage': target_stage_name,
                    'config': stage_config,
                    'params': results
                })

            return {
                'optimize_stage_results': optimization_results
            }

    @event('execute_batched_optimization')
    async def collect_optimized_batch_sizes(
        self,
        optimize_stage_results: List[Dict[str, Any]]=[],
        optimize_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]] = {},
        optimize_stage_has_multiple_workers: bool = False
    ):

        optimized_batch_sizes = []
        for optimization_result in optimize_stage_results:
            optimized_config: Config = optimization_result.get('config')
            optimized_batch_sizes.append(optimized_config.batch_size)

        optimized_batch_size = sum(optimized_batch_sizes)

        if optimize_stage_has_multiple_workers:
        
            stage_cpu_monitor: CPUMonitor = optimize_stage_monitors.get('cpu')
            stage_memory_monitor: MemoryMonitor = optimize_stage_monitors.get('memory')

            optimized_batch_sizes = []
            for optimization_result in optimize_stage_results:
                monitors: Dict[str, MonitorResults] = optimization_result.get('monitoring', {})
                worker_id = optimization_result.get('worker_id')

                cpu_monitor = monitors.get('cpu', {})
                memory_monitor = monitors.get('memory', {})
                
                for monitor_name, collection_stats in cpu_monitor.items():
                    stage_cpu_monitor.worker_metrics[worker_id][monitor_name] = collection_stats
                    stage_cpu_monitor.collected[monitor_name].extend(collection_stats)

                for monitor_name, collection_stats in memory_monitor.items():
                    stage_memory_monitor.worker_metrics[worker_id][monitor_name] = collection_stats

            stage_cpu_monitor.aggregate_worker_stats()
            stage_memory_monitor.aggregate_worker_stats()

        main_monitor_name = f'{self.name}.main'

        await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
        await stage_memory_monitor.stop_background_monitor(main_monitor_name)

        stage_cpu_monitor.close()
        stage_memory_monitor.close()

        stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
        stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]

        return {
            'optimize_stage_batch_size': optimized_batch_size,
            'optimize_stage_monitors': {
                'cpu': stage_cpu_monitor,
                'memory': stage_memory_monitor
            }
        }

    @context('collect_optimized_batch_sizes')
    async def set_optimized_configs(
        self,
        optimize_stage_results: List[Any]=[],
        optimize_stage_candidates: Dict[str, Execute]={},
        optimize_stage_batch_size: int=0
    ):
       
        stage_optimzations = {}
        stage_context = defaultdict(list)
        optimized_stages = {}
        optimzied_hooks = defaultdict(list)
        optimized_configs: Dict[str, Any] = {}
        stages_setup_by = {}

        for optimization_result in optimize_stage_results:
            
            stage_name = optimization_result.get('stage')
            optimized_config: Config = optimization_result.get('config')

            stage = optimize_stage_candidates.get(stage_name)
            optimized_configs[stage.name] = optimized_config
            stages_setup_by[stage.name] = stage.context['execute_stage_setup_by']

            for hook in stage.dispatcher.actions_and_tasks.values():
                if hook.source.session:
                    hook.source.session.pool.size = optimize_stage_batch_size
                    hook.source.session.sem = asyncio.Semaphore(optimize_stage_batch_size)
                    hook.source.session.pool.connections = []
                    hook.source.session.pool.create_pool()

                    optimzied_hooks[hook.source.stage].append(hook)

            stage.context['execute_stage_setup_hooks'] = list(stage.dispatcher.actions_and_tasks.values())
            
            pipeline_context = optimization_result.get('context', {})
            for context_key, context_value in pipeline_context.items():
                stage_context[context_key].append(context_value)

            stage_optimzations[stage_name] = optimize_stage_batch_size

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Stage - {stage_name} - configured to use optimized batch size of - {optimize_stage_batch_size} - VUs')

            optimized_config.optimized = True
            stage.optimized = True

            optimized_stages[stage.name] = stage

        return {
            'optimize_stage_optimized_hooks': optimzied_hooks,
            'optimize_stage_optimized_configs': optimized_configs,
            'execute_stage_setup_by': stages_setup_by,
            'optimize_stage_optimzations': stage_optimzations,
            'optimize_stage_context': stage_context
        }

    @context('set_optimized_configs')
    async def complete_optimization(
        self,
        optimize_stage_stage_names: str=None,
        optimize_stage_results: List[Any]=[],
        optimize_stage_workers_map: Dict[str, Execute]={},
        optimize_stage_candidates: Dict[str, Execute]={},
        optimize_stage_optimzations: Dict[str, Union[int, float]]={},
        optimize_stage_context: Dict[str, Any]={},
        optimize_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={}
    ):
        self.context[self.name] = optimize_stage_context

        for stage in optimize_stage_candidates.values():
            stage.workers = optimize_stage_workers_map.get(stage.name)

        self.optimization_execution_time = round(time.monotonic() - self.optimization_execution_time_start)

        optimized_batch_sizes = ', '.join([
            f'{stage_name}: {optimized_batch_size}' for stage_name, optimized_batch_size in optimize_stage_optimzations.items()
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Optimization complete for stages - {optimize_stage_stage_names} - over - {self.optimization_execution_time} - seconds')
        await self.logger.spinner.set_default_message(f'Optimized - batch sizes for stages - {optimized_batch_sizes} - over {self.optimization_execution_time} seconds')
        
        return {
            'optimize_stage_optimized_params': [
                result.get('params') for result in optimize_stage_results
            ],
            'optimize_stage_monitors': {
                self.name: {
                    **optimize_stage_monitors
                }
            }
        }

    @event('complete_optimization')
    async def complete(self):
        await self.executor.shutdown()




import asyncio
import re
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.timeouts import Timeouts
from typing import Dict, Any, Coroutine


class BaseParser:

    def __init__(
        self, 
        name: str,
        config: Config,
        parser_type: RequestTypes,
        options: Dict[str, Any]={}
    ) -> None:
        
        self._loop: asyncio.AbstractEventLoop = None
        self._name_pattern = re.compile('[^0-9a-zA-Z]+')
        self.name = name
        self.config = config
        self.timeouts = Timeouts(
            connect_timeout=config.connect_timeout,
            total_timeout=config.request_timeout
        )
        
        self.parser_type = parser_type
        self.options = options

    async def parse(self, action_data: Dict[str, Any]) -> Coroutine[Any, Any, ActionHook]:
        raise NotImplementedError('Parse method is not implemented for base Parser class.')
    



from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    StrictBool,
    AnyHttpUrl
)
from typing import List, Dict, Optional, Union, Any


class PlaywrightOptionsValidator(BaseModel):
    event: Optional[StrictStr]
    option: Any
    is_checked: Optional[StrictBool]
    timeout: Optional[Union[StrictInt, StrictFloat]]
    extra: Optional[Dict[str, Any]]
    switch_by: Optional[StrictStr]


class PlaywrightInputValidator(BaseModel):
    key: Optional[StrictStr]
    text: Optional[StrictStr]
    expression: Optional[StrictStr]
    args: Optional[List[Any]]
    filepath: Optional[StrictStr]
    file: Optional[StrictStr]
    path: Optional[StrictStr]
    option: Optional[Union[StrictStr, StrictBool, StrictInt, StrictFloat]]
    by_label: StrictBool=False
    by_value: StrictBool=False


class PlaywrightURLValidator(BaseModel):
    location: Optional[AnyHttpUrl]
    headers: Dict[str, str]


class PlaywrightPageValidator(BaseModel):
    selector: Optional[StrictStr]
    attribute: Optional[StrictStr]
    x_coordinate: Optional[Union[StrictInt, StrictFloat]]
    y_coordinate: Optional[Union[StrictInt, StrictFloat]]
    frame: Optional[StrictInt]


class PlaywrightActionTag(BaseModel):
    name: StrictStr
    value: StrictStr


class PlaywrightActionValidator(BaseModel):
    engine: StrictStr
    name: StrictStr
    command: StrictStr
    page: Optional[PlaywrightPageValidator]
    url: Optional[PlaywrightURLValidator]
    input: Optional[PlaywrightInputValidator]
    options: Optional[PlaywrightOptionsValidator]
    weight: Optional[Union[StrictInt, StrictFloat]]
    order: Optional[StrictInt]
    user: Optional[StrictStr]
    tags: List[PlaywrightActionTag]=[]




from __future__ import annotations
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    load_string,
    pack_string
)


class TXTRecordData(RecordData):
    '''A record'''

    def __init__(self, data: str):
        super().__init__(
            RecordType.TXT,
            data=data
        )

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, TXTRecordData]:
        
        _, text = load_string(
            data, 
            cursor_position
        )

        return cursor_position + size, TXTRecordData(
            text.decode()
        )

    def dump(
        self, 
        names: Dict[str, int], 
        offset: int
    ) -> Iterable[bytes]:
        yield pack_string(self.data)




import time
import threading
import os
import dill
from collections import defaultdict
from typing import Any, Dict, List
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.logging import (
    HedraLogger,
    LoggerTypes
)
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.reporting.processed_result.processed_results_group import ProcessedResultsGroup
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes


def process_results_batch(config: Dict[str, Any]):
    
    import warnings
    warnings.simplefilter("ignore")

    from hedra.logging import (
        logging_manager
    )
    
    graph_name = config.get('graph_name')
    graph_id = config.get('graph_id')
    logfiles_directory = config.get('logfiles_directory')
    log_level = config.get('log_level')
    source_stage_name = config.get('source_stage_name')
    source_stage_id = config.get('source_stage_id')
    enable_unstable_features = config.get('enable_unstable_features', False)
    worker_id = config.get('worker_id')

    cpu_monitor = CPUMonitor()
    memory_monitor = MemoryMonitor()

    cpu_monitor.stage_type = StageTypes.ANALYZE
    memory_monitor.stage_type = StageTypes.ANALYZE

    monitor_name = f'{source_stage_name}.worker'
    cpu_monitor.start_background_monitor_sync(monitor_name)
    memory_monitor.start_background_monitor_sync(monitor_name)

    active_flags[FlagTypes.UNSTABLE_FEATURE] = enable_unstable_features

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM,
        LoggerTypes.SPINNER
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory

    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    stage_name = config.get('analyze_stage_name')
    results_batch: List[BaseResult] = config.get('analyze_stage_batched_results', [])

    try:

        events =  defaultdict(ProcessedResultsGroup)

        logger = HedraLogger()
        logger.initialize()
        logger.filesystem.sync.create_logfile('hedra.core.log')
        logger.filesystem.sync.create_logfile('hedra.reporting.log')
        
        logger.filesystem.sync['hedra.core'].info(f'{metadata_string} - Initializing results aggregation')

        start = time.monotonic()

        for result in results_batch:
            stage_result: BaseResult = dill.loads(result)
            events[stage_result.name].add(
                    stage_name,
                    stage_result,
                )

        for events_stage_name, events_group in events.items():  

            logger.filesystem.sync['hedra.reporting'].debug(
                f'{metadata_string} - Events group - {events_group.events_group_id} - created for Stage - {events_stage_name} - Processed - {events_group.total}'
            )

            events_group.calculate_stats()

        elapsed = time.monotonic() - start

        logger.filesystem.sync['hedra.core'].info(f'{metadata_string} - Results aggregation complete - Took: {round(elapsed, 2)} seconds')


        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        return {
            'worker_id': worker_id,
            'events': events,
            'monitoring': {
                'cpu': cpu_monitor.collected,
                'memory': memory_monitor.collected
            }
        }

    except BrokenPipeError:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise ProcessKilledError()
    
    except RuntimeError:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise ProcessKilledError()

    except Exception as e:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise e



from hedra.core.graphs.stages.types.stage_types import StageTypes
from .common.transtition_metadata import TransitionMetadata


local_transitions = {

        # State: Idle
        (StageTypes.IDLE, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=True, is_valid=True),
        (StageTypes.IDLE, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=True, is_valid=True),
        (StageTypes.IDLE, StageTypes.SETUP):  TransitionMetadata(allow_multiple_edges=True, is_valid=True),
        (StageTypes.IDLE, StageTypes.OPTIMIZE):  TransitionMetadata(allow_multiple_edges=True, is_valid=False),
        (StageTypes.IDLE, StageTypes.EXECUTE):   TransitionMetadata(allow_multiple_edges=True, is_valid=False),
        (StageTypes.IDLE, StageTypes.ANALYZE):  TransitionMetadata(allow_multiple_edges=True, is_valid=False),
        (StageTypes.IDLE, StageTypes.SUBMIT):  TransitionMetadata(allow_multiple_edges=True, is_valid=False),
        (StageTypes.IDLE, StageTypes.COMPLETE):  TransitionMetadata(allow_multiple_edges=True, is_valid=False),
        (StageTypes.IDLE, StageTypes.ERROR):  TransitionMetadata(allow_multiple_edges=True, is_valid=False),

        # State: Setup
        (StageTypes.SETUP, StageTypes.SETUP):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SETUP, StageTypes.ACT):  TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SETUP, StageTypes.IDLE):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SETUP, StageTypes.OPTIMIZE):  TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SETUP, StageTypes.EXECUTE):  TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SETUP, StageTypes.ANALYZE):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SETUP, StageTypes.SUBMIT):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SETUP, StageTypes.COMPLETE):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SETUP, StageTypes.ERROR):  TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Optimize
        (StageTypes.OPTIMIZE, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.OPTIMIZE, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.OPTIMIZE, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.OPTIMIZE, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),


        # State: Act
        (StageTypes.ACT, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ACT, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ACT, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Execute
        (StageTypes.EXECUTE, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.EXECUTE, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.EXECUTE, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.EXECUTE, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.EXECUTE, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.EXECUTE, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.EXECUTE, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.EXECUTE, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.EXECUTE, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Analyze
        (StageTypes.ANALYZE, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ANALYZE, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.ANALYZE, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ANALYZE, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Submit
        (StageTypes.SUBMIT, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SUBMIT, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.SUBMIT, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.SUBMIT, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Complete
        (StageTypes.COMPLETE, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=True),
        (StageTypes.COMPLETE, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),

        # State: Error
        (StageTypes.ERROR, StageTypes.ERROR): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.ACT): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.IDLE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.SETUP): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.OPTIMIZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.EXECUTE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.ANALYZE): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.SUBMIT): TransitionMetadata(allow_multiple_edges=False, is_valid=False),
        (StageTypes.ERROR, StageTypes.COMPLETE): TransitionMetadata(allow_multiple_edges=False, is_valid=False)
    }


# # This is an ugly patch for: https://github.com/aio-libs/aiopg/issues/837
# import selectors  # isort:skip # noqa: F401

# selectors._PollLikeSelector.modify = (  # type: ignore
#     selectors._BaseSelectorImpl.modify  # type: ignore
# )
import asyncio
import uuid
from typing import (
    List, 
    Dict, 
    Any, 
    Union,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .postgres_connector_config import PostgresConnectorConfig


try:
    import sqlalchemy
    from sqlalchemy.engine.result import Result as SQLResult
    from sqlalchemy.ext.asyncio import create_async_engine
    from sqlalchemy.ext.asyncio.engine import (
        AsyncEngine, 
        AsyncConnection
    )

    has_connector = True

except Exception:
    SQLResult = object
    sqlalchemy = object
    AsyncEngine = object
    AsyncConnection = object
    AsyncTransaction = object
    create_async_engine: lambda *args, **kwargs: None
    has_connector = False


class PostgresConnection:
    connection_type=ConnectorType.Postgres

    def __init__(
        self, 
        config: PostgresConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        
        self.host = config.host
        self.database = config.database
        self.username = config.username
        self.password = config.password
        self.stage = stage
        self.parser_config = parser_config

        self.table_name = config.table_name
        
        self._engine: Union[AsyncEngine, None] = None
        self.metadata = sqlalchemy.MetaData()

        self._table = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        
        self.parser = Parser()
        self.sql_type = 'Postgresql'

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to {self.sql_type} instance at - {self.host} - Database: {self.database}')

        connection_uri = 'postgresql+asyncpg://'

        if self.username and self.password:
            connection_uri = f'{connection_uri}{self.username}:{self.password}@'

        self._engine: AsyncEngine = await create_async_engine(
            f'{connection_uri}{self.host}/{self.database}',
            echo=False
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to {self.sql_type} instance at - {self.host} - Database: {self.database}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> ExecuteStageSummaryValidator:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        if self._table is None:
            
            self._table = sqlalchemy.Table(
                self.table_name,
                self.metadata,
                autoload_with=self._engine
            )
            
        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection
            
            results: SQLResult = await connection.execute(
                self._table.select(**options)
            )

            return [
                {
                    column: value for column, value in row._mapping.items()
                } for row in results.fetchall()
            ]

    async def close(self):
        pass


import asyncio
import json
import psutil
import signal
import uuid
from typing import List, Dict
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .google_cloud_storage_config import GoogleCloudStorageConfig


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


try:

    from google.cloud import storage
    has_connector = True

except Exception:
    storage = None
    has_connector = False

class GoogleCloudStorage:

    def __init__(self, config: GoogleCloudStorageConfig) -> None:
        self.service_account_json_path = config.service_account_json_path

        self.bucket_namespace = config.bucket_namespace
        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.streams_bucket_name = config.streams_bucket

        self.experiments_bucket_name = config.experiments_bucket
        self.variants_bucket_name = f'{config.experiments_bucket}_variants'
        self.mutations_bucket_name = f'{config.experiments_bucket}_mutations'
        
        self.shared_metrics_bucket_name = f'{config.metrics_bucket}_shared'
        self.errors_bucket_name = f'{config.metrics_bucket}_errors'
        self.custom_metrics_bucket_name = f'{config.metrics_bucket}_custom'

        self.session_system_metrics_bucket_name = f'{config.system_metrics_bucket}_session'
        self.stage_system_metrics_bucket_name = f'{config.system_metrics_bucket}_stage'

        self.credentials = None
        self.client = None

        self._events_bucket = None
        self._streams_bucket = None

        self._experiments_bucket = None
        self._variants_bucket = None
        self._mutations_bucket = None

        self._shared_metrics_bucket = None
        self._metrics_bucket = None
        self._errors_bucket = None
        self._custom_metrics_bucket = None
        self._session_system_metrics_bucket = None
        self._stage_system_metrics_bucket = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        self.client = storage.Client.from_service_account_json(self.service_account_json_path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Session System Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.session_system_metrics_bucket_name} if not exists')

            self._session_system_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.session_system_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Session System Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.session_system_metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Session System Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.session_system_metrics_bucket_name}')

            self._session_system_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.session_system_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Session System Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.session_system_metrics_bucket_name}')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.session_system_metrics_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics Set - {metrics_set.name}:{metrics_set.group}')

            blob = await self._loop.run_in_executor(
                self._executor,
                self._streams_bucket.blob,
                f'{metrics_set.name}_{metrics_set.group}_{self.session_uuid}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(metrics_set.record)
            )
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Stage System Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.stage_system_metrics_bucket_name} if not exists')

            self._stage_system_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.stage_system_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Stage System Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.stage_system_metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Stage System Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.stage_system_metrics_bucket_name}')

            self._stage_system_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.stage_system_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Stage System Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.stage_system_metrics_bucket_name}')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.stage_system_metrics_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics Set - {metrics_set.name}:{metrics_set.group}')

            blob = await self._loop.run_in_executor(
                self._executor,
                self._streams_bucket.blob,
                f'{metrics_set.name}_{metrics_set.group}_{self.session_uuid}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(metrics_set.record)
            )

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Streams bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.streams_bucket_name} if not exists')

            self._streams_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.streams_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Streams bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.streams_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Streams bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.streams_bucket_name}')

            self._streams_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.streams_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Streams bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.streams_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to - Namespace: {self.bucket_namespace} - Bucket: {self.streams_bucket_name}')
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():
                blob = await self._loop.run_in_executor(
                    self._executor,
                    self._streams_bucket.blob,
                    f'{stage_name}_{group_name}_{self.session_uuid}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        'name': f'{stage_name}_streams',
                        'stage': stage_name,
                        'group': group_name,
                        **group
                    })
                )

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Experiments bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name} if not exists')

            self._experiments_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.experiments_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Experiments bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Experiments bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name}')

            self._experiments_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.experiments_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Experiments bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name}')
        for experiment in experiment_metrics.experiment_summaries:

            experiment_id = uuid.uuid4()

            blob = await self._loop.run_in_executor(
                self._executor,
                self._events_bucket.blob,
                f'{experiment.experiment_name}_{self.session_uuid}_{experiment_id}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(experiment.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to - Namespace: {self.bucket_namespace} - Bucket: {self.experiments_bucket_name}')
    
    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Variants bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name} if not exists')

            self._variants_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.variants_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Variants bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Variants bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name}')

            self._variants_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.variants_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Variants bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name}')
        for variant in experiment_metrics.variant_summaries:

            variant_id = uuid.uuid4()

            blob = await self._loop.run_in_executor(
                self._executor,
                self._events_bucket.blob,
                f'{variant.variant_name}_{self.session_uuid}_{variant_id}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(variant.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to - Namespace: {self.bucket_namespace} - Bucket: {self.variants_bucket_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Mutations bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name} if not exists')

            self._mutations_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.mutations_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Mutations bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Mutations bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name}')

            self._mutations_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.mutations_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Mutations bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name}')
        for mutation in experiment_metrics.mutation_summaries:

            mutation_id = uuid.uuid4()

            blob = await self._loop.run_in_executor(
                self._executor,
                self._events_bucket.blob,
                f'{mutation.mutation_name}_{self.session_uuid}_{mutation_id}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(mutation.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to - Namespace: {self.bucket_namespace} - Bucket: {self.mutations_bucket_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name} if not exists')

            self._events_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.events_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Events bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

            self._events_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.events_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Events bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')
        for event in events:
            blob = await self._loop.run_in_executor(
                self._executor,
                self._events_bucket.blob,
                f'{event.name}_{self.session_uuid}_{event.event_id}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps(event.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Namespace: {self.bucket_namespace} - Bucket: {self.events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name} if not exists')

            self._shared_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.shared_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Shared Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')

            self._shared_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.shared_metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Shared Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            blob = await self._loop.run_in_executor(
                self._executor,
                self.metrics_bucket.blob,
                f'{metrics_set.name}_shared_{self.session_uuid}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.shared_metrics_bucket_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name} if not exists')

            self._metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_{self.metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')

            self._metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_{self.metrics_bucket_name}'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                blob = await self._loop.run_in_executor(
                    self._executor,
                    self._metrics_bucket.blob,
                    f'{metrics_set.name}_{group_name}_{self.session_uuid}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        **group.record,
                        'group': group_name
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name} if not exists')

            self._custom_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                self.custom_metrics_bucket_name
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name}')

            self._custom_metrics_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                self.custom_metrics_bucket_name
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Custom Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name}')
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            blob = await self._loop.run_in_executor(
                self._executor,
                self._metrics_bucket.blob,
                f'{metrics_set.name}_custom_{self.session_uuid}'
            )

            await self._loop.run_in_executor(
                self._executor,
                blob.upload_from_string,
                json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    **{
                        custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                    }
                })
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.custom_metrics_bucket_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Errors Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name} if not exists')

            self._errors_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.get_bucket,
                f'{self.bucket_namespace}_errors'
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Errors Metrics bucket at - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')
        
        except Exception:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Setting Error Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')

            self._errors_bucket = await self._loop.run_in_executor(
                self._executor,
                self.client.create_bucket,
                f'{self.bucket_namespace}_errors'
            )
            
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Set Errors Metrics bucket as - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                
                blob = await self._loop.run_in_executor(
                    self._executor,
                    self.metrics_bucket.blob,
                    f'{metrics_set.name}_errors_{self.session_uuid}'
                )

                await self._loop.run_in_executor(
                    self._executor,
                    blob.upload_from_string,
                    json.dumps({
                        'metric_name': metrics_set.name,
                        'metric_stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to - Namespace: {self.bucket_namespace} - Bucket: {self.errors_bucket_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing Google Cloud connection')
        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )

        self._executor.shutdown(wait=False, cancel_futures=True)

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed Google Cloud connection')


import asyncio
import json
import time
import uuid
from typing import (
    List, 
    Dict, 
    Any,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .redis_connector_config import RedisConnectorConfig


try:

    import aioredis
    
    has_connector = True

except Exception:
    aioredis = None
    has_connector = True


class RedisConnector:
    connector_type=ConnectorType.Redis

    def __init__(
        self, 
        config: RedisConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.host = config.host
        self.base = 'rediss' if config.secure else 'redis'
        self.username = config.username
        self.password = config.password
        self.database = config.database
        self.channel = config.channel
        self.stage = stage
        self.parser_config = parser_config

        self.channel_type = config.channel_type
        self.connection = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.parser = Parser()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Redis instance at - {self.base}://{self.host} - Database: {self.database}')
        
        self.connection = await aioredis.from_url(
            f'{self.base}://{self.host}',
            username=self.username,
            password=self.password,
            db=self.database
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Redis instance at - {self.base}://{self.host} - Database: {self.database}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        if self.channel:
            async with self.connection.client() as connection:
                keys = await connection.keys()
                return await asyncio.gather(*[
                    asyncio.create_task(
                        connection.get(key) for key in keys
                    )
                ])
            
        else:
            subscriber = self.connection.pubsub()
            await subscriber.subscribe(self.channel)

            limit = options.get('limit')
            timeout = options.get('timeout', 1)

            if limit is None and timeout is None:
                raise Exception('A limit or timeout must be provided.')
            
            if limit:

                async with subscriber as subscription:
                    results: List[bytes] = await asyncio.gather(*[
                        asyncio.create_task(subscription.get_message(
                            ignore_subscribe_messages=True
                        )) for _ in range(limit)
                    ])
                
            else:
                async with subscriber as subscription:
                    elapsed = 0
                    start = time.time()

                    results: List[bytes] = []

                    while elapsed < timeout:
                        results.append(
                            await subscription.get_message(
                                ignore_subscribe_messages=True
                            )
                        )

                        elapsed = time.time() - start

            return [
                json.loads(result) for result in results
            ]  
    
    async def close(self):
        await self.connection.close()



import uuid
import time
import asyncio
import psutil
import math
from typing import List
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.default_persona.default_persona import DefaultPersona
from hedra.core.personas.types.types import PersonaTypes
from hedra.versioning.flags.types.unstable.flag import unstable


@unstable
class ApproximateDistributionPersona(DefaultPersona):

    __slots__ = (
        'metadata_string',
        'persona_id',
        'logger',
        'type',
        'workers',
        'actions',
        '_hooks',
        'batch',
        '_stream',
        'total_time',
        'duration',
        'total_actions',
        'total_elapsed',
        'start',
        'end',
        'completed_actions',
        'pending_actions',
        'completed_time',
        'run_timer',
        'actions_count',
        'graceful_stop',
        'is_timed',
        '_stream_thread',
        '_loop',
        'current_action_idx',
        'optimized_params',
        '_executor',
        'stream',
        'execution_metrics',
        'stream_reporter_configs',
        'stream_reporters',
        'stage_name',
        'distribution',
        'collection_interval'
    )    

    def __init__(self, config: Config):
        super().__init__(config)

        self.distribution: List[float] = config.experiment.get('distribution')
        self.collection_interval = config.experiment.get('interval_duration')
        self.persona_id = str(uuid.uuid4())
        self.type = PersonaTypes.APPROXIMATE_DISTRIBUTION

    async def generator(self, total_time):
        elapsed = 0
        idx = 0
        action_idx = 0
        current_interval_idx = 0

        distribution = self.distribution

        intervals_count = len(distribution)
        time_interval_amount = math.floor(total_time/intervals_count)
        
        interval_periods = [
             time_interval_amount for _ in distribution
        ]

        remainder_period = total_time%intervals_count
        if remainder_period > 0:
            interval_periods[-1] += remainder_period
             

        max_pool_size = math.ceil(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)
        
        interval_period = interval_periods[current_interval_idx]
        interval_time_limit = interval_period
        distribution_batch_size = distribution[current_interval_idx]

        for action in self._hooks:
            await action.session.set_pool(distribution_batch_size)

        start = time.time()
        while elapsed < total_time:
            yield action_idx
            
            await asyncio.sleep(0)
            elapsed = time.time() - start
            idx += 1

            if idx%self.batch.size == 0:
                await asyncio.sleep(self.batch.interval)

            action_idx = (action_idx + 1)%self.actions_count
            
            if self._hooks[action_idx].session.active > max_pool_size:
                try:
                    max_wait = total_time - elapsed
                    await asyncio.wait_for(
                        self._hooks[action_idx].session.wait_for_active_threshold(),
                        timeout=max_wait
                    )
                except asyncio.TimeoutError:
                    pass

            if elapsed >= interval_time_limit and elapsed < total_time:
                
                current_interval_idx += 1
                interval_time_limit += interval_periods[current_interval_idx]
                distribution_batch_size = distribution[current_interval_idx]
       
                for action in self._hooks:
                    await action.session.set_pool(distribution_batch_size)



import functools
from typing import Tuple
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from .validator import SaveHookValidator


@registrar(HookType.SAVE)
def save(
    *names: Tuple[str, ...], 
    save_path: str=None,
    order: int=1,
    skip: bool=False
):
    SaveHookValidator(
        names=names,
        save_path=save_path,
        order=order,
        skip=skip
    )
    
    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper



import click
import atexit
import warnings
import uvloop
from multiprocessing import current_process, active_children

from .base import CLI
uvloop.install()
warnings.simplefilter("ignore")

@click.group(cls=CLI)
def run():

    def stop_processes_at_exit():
        child_processes = active_children()
        for child in child_processes:
            child.close()

        process = current_process()
        if process:
            process.close()

        else:
            print('\n')
 
    atexit.register(stop_processes_at_exit)



import asyncio
import psutil
import signal
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .snowflake_config import SnowflakeConfig


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass
    


try:
    import sqlalchemy
    from snowflake.sqlalchemy import URL
    from sqlalchemy import create_engine
    from sqlalchemy.schema import CreateTable
    
    has_connector = True

except Exception:
    snowflake = None
    has_connector = False

class Snowflake:

    def __init__(self, config: SnowflakeConfig) -> None:
        self.username = config.username
        self.password = config.password
        self.organization_id = config.organization_id
        self.account_id = config.account_id
        self.private_key = config.private_key
        self.warehouse = config.warehouse
        self.database = config.database
        self.schema = config.database_schema

        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.streams_table_name = config.streams_table

        self.experiments_table_name = config.experiments_table
        self.variants_table_name = f'{config.experiments_table}_variants'
        self.mutations_table_name = f'{config.experiments_table}_mutations'

        self.shared_metrics_table_name = f'{config.metrics_table}_shared'
        self.errors_table_name = f'{config.metrics_table}_errors'
        self.custom_metrics_table_name = f'{config.metrics_table}_custom'


        self.session_system_metrics_table_name = f'{config.system_metrics_table}_session'
        self.stage_system_metrics_table_name = f'{config.system_metrics_table}_stage'

        self.connect_timeout = config.connect_timeout
        
        self.metadata = sqlalchemy.MetaData()
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self._engine = None
        self._connection = None

        self._events_table = None
        self._metrics_table = None
        self._streams_table = None

        self._experiments_table = None
        self._variants_table = None
        self._mutations_table = None

        self._shared_metrics_table = None
        self._custom_metrics_table = None
        self._errors_table = None

        self._session_system_metrics_table = None
        self._stage_system_metrics_table = None

        self.metric_types_map = {
            MetricType.COUNT: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.Integer
            ),
            MetricType.DISTRIBUTION: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
            MetricType.SAMPLE: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
            MetricType.RATE: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
        }

        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        try:

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop
                    )
                )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')
            self._engine = await self._loop.run_in_executor(
                self._executor,
                create_engine,
                URL(
                    user=self.username,
                    password=self.password,
                    account=self.account_id,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema
                )
            
            )

            self._connection = await asyncio.wait_for(
                self._loop.run_in_executor(
                    self._executor,
                    self._engine.connect
                ),
                timeout=self.connect_timeout
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')

        except asyncio.TimeoutError:
            raise Exception('Err. - Connection to Snowflake timed out - check your account id, username, and password.')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Table - {self.session_system_metrics_table_name}')

        if self._session_system_metrics_table is None:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Session System Metrics table - {self.session_system_metrics_table_name} - if not exists')

            session_system_metrics_table = sqlalchemy.Table(
                self.session_system_metrics_table_name,
                self.metadata,
                sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('median', sqlalchemy.FLOAT),
                sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
            )

            for quantile in SystemMetricsSet.quantiles:
                session_system_metrics_table.append_column(
                    sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                )
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                CreateTable(
                    session_system_metrics_table, 
                    if_not_exists=True
                )
            )

            self._session_system_metrics_table = session_system_metrics_table
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Session System Metrics table - {self.session_system_metrics_table_name}')

        rows: List[SessionMetricsCollection] = []
            
        for metrics_set in system_metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')

            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                rows.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                rows.append(monitor_metrics)

        for metrics_set in rows:
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._session_system_metrics_table.insert(values=metrics_set.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Table - {self.stage_system_metrics_table_name}')
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Table - {self.stage_system_metrics_table_name}')

        if self._stage_system_metrics_table is None:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stage System Metrics table - {self.stage_system_metrics_table_name} - if not exists')

            stage_system_metrics_table = sqlalchemy.Table(
                self.stage_system_metrics_table_name,
                self.metadata,
                sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('median', sqlalchemy.FLOAT),
                sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
            )

            for quantile in SystemMetricsSet.quantiles:
                stage_system_metrics_table.append_column(
                    sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                )
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                CreateTable(
                    stage_system_metrics_table, 
                    if_not_exists=True
                )
            )

            self._stage_system_metrics_table = stage_system_metrics_table
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Stage System Metrics table - {self.stage_system_metrics_table_name}')

        rows: List[SystemMetricsCollection] = []
            
        for metrics_set in system_metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    rows.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    rows.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    rows.append(stage_mb_per_vu_metrics)

        for metrics_set in rows:
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._stage_system_metrics_table.insert(values=metrics_set.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Table - {self.stage_system_metrics_table_name}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Table - {self.streams_table_name}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')

            if self._streams_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Streams table - {self.streams_table_name} - if not exists')

                streams_table = sqlalchemy.Table(
                    self.streams_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('median', sqlalchemy.FLOAT),
                    sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                    sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                    sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                    sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                )

                for quantile in stream.quantiles:
                    streams_table.append_column(
                        sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                    )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(
                        streams_table, 
                        if_not_exists=True
                    )
                )

                self._streams_table = streams_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Streams table - {self.streams_table_name}')

        for group_name, group in stream.grouped.items():
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._streams_table.insert(values={
                    **group,
                    'group': group_name,
                    'stage': stage_name,
                    'name': f'{stage_name}_streams'
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Table - {self.streams_table_name}')

    async def submit_experiments(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Table - {self.experiments_table_name}')
    
        for experiment in experiments_metrics.experiment_summaries:

            if self._experiments_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Experiments table - {self.experiments_table_name} - if not exists')
                experiments_table = sqlalchemy.Table(
                    self.experiments_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('experiment_name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('experiment_randomized', sqlalchemy.Boolean),
                    sqlalchemy.Column('experiment_completed', sqlalchemy.BIGINT),
                    sqlalchemy.Column('experiment_succeeded', sqlalchemy.BIGINT),
                    sqlalchemy.Column('experiment_failed', sqlalchemy.BIGINT),
                    sqlalchemy.Column('experiment_median_aps', sqlalchemy.FLOAT),
                )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(experiments_table, if_not_exists=True)
                )

                self._experiments_table = experiments_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Experiments table - {self.experiments_table_name}')
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._experiments_table.insert(values=experiment.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Table - {self.experiments_table_name}')

    async def submit_variants(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Table - {self.variants_table_name}')
    
        for variant in experiments_metrics.variant_summaries:

            if self._variants_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Variants table - {self.variants_table_name} - if not exists')
                variants_table = sqlalchemy.Table(
                    self.variants_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('variant_name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('variant_experiment', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('variant_weight', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_distribution', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('variant_distribution_interval', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_completed', sqlalchemy.BIGINT),
                    sqlalchemy.Column('variant_succeeded', sqlalchemy.BIGINT),
                    sqlalchemy.Column('variant_failed', sqlalchemy.BIGINT),
                    sqlalchemy.Column('variant_actions_per_second', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_ratio_completed', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_ratio_succeeded', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_ratio_failed', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variant_ratio_aps', sqlalchemy.FLOAT),
                )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(variants_table, if_not_exists=True)
                )

                self._variants_table = variants_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Variants table - {self.variants_table_name}')
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._variants_table.insert(values=variant.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Table - {self.variants_table_name}')

    async def submit_mutations(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Table - {self.mutations_table_name}')
    
        for variant in experiments_metrics.variant_summaries:

            if self._mutations_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Mutations table - {self.mutations_table_name} - if not exists')
                mutations_table = sqlalchemy.Table(
                    self.mutations_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('mutation_name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('mutation_experiment_name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('mutation_variant_name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('mutation_chance', sqlalchemy.FLOAT),
                    sqlalchemy.Column('mutation_targets', sqlalchemy.VARCHAR(8192)),
                    sqlalchemy.Column('mutation_type', sqlalchemy.VARCHAR(255)),
                )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(mutations_table, if_not_exists=True)
                )

                self._mutations_table = mutations_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Mutations table - {self.mutations_table_name}')
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._mutations_table.insert(values=variant.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Table - {self.mutations_table_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name}')
    
        for event in events:

            if self._events_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - if not exists')
                events_table = sqlalchemy.Table(
                    self.events_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('time', sqlalchemy.Float),
                    sqlalchemy.Column('succeeded', sqlalchemy.Boolean),
                )
                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(events_table, if_not_exists=True)
                )

                self._events_table = events_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Events table - {self.events_table_name}')
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._events_table.insert(values=event.record)
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table - {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name}')

        if self._shared_metrics_table is None:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - if not exists')

            shared_metrics_table = sqlalchemy.Table(
                self.shared_metrics_table_name,
                self.metadata,
                sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('group', sqlalchemy.TEXT),
                sqlalchemy.Column('total', sqlalchemy.BIGINT),
                sqlalchemy.Column('succeeded', sqlalchemy.BIGINT),
                sqlalchemy.Column('failed', sqlalchemy.BIGINT),
                sqlalchemy.Column('actions_per_second', sqlalchemy.FLOAT)
            )

            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                CreateTable(shared_metrics_table, if_not_exists=True)
            )

            self._shared_metrics_table = shared_metrics_table
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Shared Metrics table - {self.shared_metrics_table_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._shared_metrics_table.insert(values={
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table - {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._metrics_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - if not exists')

                metrics_table = sqlalchemy.Table(
                    self.metrics_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('median', sqlalchemy.FLOAT),
                    sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                    sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                    sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                    sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                    sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                )

                for quantile in metrics_set.quantiles:
                    metrics_table.append_column(
                        sqlalchemy.Column(f'{quantile}', sqlalchemy.FLOAT)
                    )

                for custom_field_name, sql_alchemy_type in metrics_set.custom_schemas:
                    metrics_table.append_column(custom_field_name, sql_alchemy_type)  

                
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(metrics_table, if_not_exists=True)
                )

                self._metrics_table = metrics_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Metrics table - {self.metrics_table_name}')

        for group_name, group in metrics_set.groups.items():
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._metrics_table.insert(values={
                    **group.record,
                    'group': group_name
                })
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table - {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to table - {self.custom_metrics_table_name}')

        if self._custom_metrics_table is None:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics table - {self.custom_metrics_table_name} - if not exists')

            custom_metrics_table = sqlalchemy.Table(
                self.custom_metrics_table_name,
                self.metadata,
                sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                sqlalchemy.Column('group', sqlalchemy.TEXT),
            )

            for metrics_set in metrics_sets:
                for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                    custom_metrics_table.append_column(
                        self.metric_types_map.get(
                            custom_metric.metric_type,
                            lambda field_name: sqlalchemy.Column(
                                field_name, 
                                sqlalchemy.FLOAT
                            )
                        )(custom_metric_name)
                    )
            
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                CreateTable(custom_metrics_table, if_not_exists=True)
            )

            self._custom_metrics_table = custom_metrics_table
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Custom Metrics table - {self.custom_metrics_table_name}')


        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                
            await self._loop.run_in_executor(
                self._executor,
                self._connection.execute,
                self._custom_metrics_table.insert(values={
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    **{
                        custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                    }
                })
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to table - {self.custom_metrics_table_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors_table is None:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics table - {self.errors_table_name} - if not exists')

                errors_table = sqlalchemy.Table(
                    self.errors_table_name,
                    self.metadata,
                    sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                    sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                    sqlalchemy.Column('stage', sqlalchemy.TEXT),
                    sqlalchemy.Column('error_message', sqlalchemy.TEXT),
                    sqlalchemy.Column('error_count', sqlalchemy.BIGINT)
                )

                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    CreateTable(errors_table, if_not_exists=True)
                )

                self._errors_table = errors_table
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Error Metrics table - {self.errors_table_name}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    self._connection.execute,
                    self._errors_table.insert(values={
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table - {self.errors_table_name}')
        
    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')

        await self._loop.run_in_executor(
            None,
            self._connection.close
        )

        self._executor.shutdown()

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closed session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Snowflake instance at - Warehouse: {self.warehouse} - Database: {self.database} - Schema: {self.schema}')


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    StrictBool,
    AnyHttpUrl,
    Json
)


from typing import List, Optional, Union


class UDPActionTag(BaseModel):
    name: StrictStr
    value: StrictStr


class UDPActionValidator(BaseModel):
    engine: StrictStr
    name: StrictStr
    url: AnyHttpUrl
    wait_for_response: StrictBool=False
    data: Optional[Union[StrictStr, Json]]
    weight: Optional[Union[StrictInt, StrictFloat]]
    order: Optional[StrictInt]
    user: Optional[StrictStr]
    tags: List[UDPActionTag]=[]


from __future__ import annotations
import networkx
from collections import defaultdict
from typing import List, Union, Dict
from hedra.core.hooks.types.base.event import BaseEvent
from hedra.core.hooks.types.base.hook import Hook, HookType
from .event_dispatch import EventDispatcher
from .get_event import get_event

class EventGraph:

    def __init__(self, hooks_by_type: Dict[HookType, Dict[str, Hook]]) -> None:
        self.hooks_by_type = hooks_by_type
        self.hooks_by_name = {}
        self.hooks_by_shortname: Dict[str, Dict[str, Hook]] = defaultdict(dict)

        for hook_type in self.hooks_by_type:
            for hook in self.hooks_by_type[hook_type].values():
                self.hooks_by_name[hook.name] = hook
                self.hooks_by_shortname[hook.stage][hook.shortname] = hook

        self.hooks_graph = networkx.DiGraph()

        self.hooks_graph.add_nodes_from([
            (
                hook_name,
                {'hook': hook}
            ) for hook_name, hook in self.hooks_by_name.items()
        ])

        self.event_hooks: List[Hook] = [
            *list(self.hooks_by_type.get(
                HookType.ACTION, 
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.CHANNEL,
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.CHECK,
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.CONDITION, 
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.CONTEXT, 
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.EVENT, 
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.LOAD,
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.METRIC,
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.SAVE,
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.TASK, 
                {}
            ).values()),
            *list(self.hooks_by_type.get(
                HookType.TRANSFORM, 
                {}
            ).values()),
        ]
        
        nodes = [self.hooks_by_name.get(node) for node in self.hooks_graph.nodes()]
        self.events: Dict[str, BaseEvent] = {}
        self.base_stages = list(set([node.stage_instance.__class__.__base__.__name__ for node in nodes]))
        self.removal_targets = []
        self.action_and_task_events = {}

    def hooks_to_events(self) -> EventGraph:
        for event_hook in self.event_hooks:
            for idx, target_hook_name in enumerate(event_hook.names): 
                target: Union[Hook, None] = self.hooks_by_name.get(target_hook_name)

                if target is None:
                    target: Hook = self.hooks_by_shortname[event_hook.stage].get(target_hook_name)
                    event_hook.names[idx] = target.name
   
                event = get_event(target, event_hook)
                
                if isinstance(target, Hook):

                    target: BaseEvent = get_event(target, target)
                    self.events[target.event_name] = target

                    self.hooks_graph.update(nodes=[(
                        target.event_name,
                        {'hook': target}
                    )])

  
                self.events[event.event_name] = event

                self.hooks_graph.update(nodes=[(
                    event.event_name,
                    {'hook': event}
                )])

            if self.events.get(event_hook.name) is None:

                event = get_event(event_hook, event_hook)

                self.events[event_hook.name] = event
                self.hooks_graph.update(nodes=[(
                    event_hook.name,
                    {'hook': event}
                )])

        return self

    def assemble_graph(self) -> EventGraph:      
        for event_hook in self.event_hooks:  
            for target_hook_name in event_hook.names: 
                target: BaseEvent = self.events.get(target_hook_name)

                if target is None:
                    target_hook: Hook = self.hooks_by_shortname[event_hook.stage].get(target_hook_name)
                    target_name = target_hook.name

                else:
                    target_name = target.event_name

                self.hooks_graph.add_edge(target_name, event_hook.name)


        return self

    def apply_graph_to_events(self) -> None:

        for event in self.events.values():

            event.execution_path = [edge for edge in networkx.bfs_layers(self.hooks_graph, event.event_name)]
            event.previous_map = [node for node in self.hooks_graph.predecessors(event.event_name)]
            event.next_map = [node for node in self.hooks_graph.successors(event.event_name)]

            for path_layer in event.execution_path:
                event.events.update({
                    event_name: self.events.get(event_name) for event_name in path_layer
                })

            hook_names = [hook.name for hook in event.source.stage_instance.hooks[event.hook_type]]

            if event.event_name in hook_names:
                hook_idx = hook_names.index(event.event_name)
                event.source.stage_instance.hooks[event.hook_type][hook_idx] = event

            else:
                event.source.stage_instance.hooks[event.hook_type].append(event)  

            if event.source.stage_instance.dispatcher is None:
                event.source.stage_instance.dispatcher = EventDispatcher()
                
            if len(event.previous_map) < 1 and event.event_name in hook_names:

                event.source.stage_instance.dispatcher.source_name = event.stage
                event.source.stage_instance.dispatcher.initial_events[event.source.stage].append(event)
            
            for layer in event.execution_path:
                for event_name in layer:
                    next_event = self.events.get(event_name)
                    event.source.stage_instance.dispatcher.add_event(next_event)

            
            event.source.stage_instance.dispatcher.add_event(event)
        
     


from .graphql.graphql_action_parser import GraphQLActionParser
from .graphql.graphql_result_parser import GraphQLResultParser
from .graphql_http2.graphql_http2_action_parser import GraphQLHTTP2ActionParser
from .graphql_http2.graphql_http2_result_parser import GraphQLHTTP2ResultParser
from .grpc.grpc_action_parser import GRPCActionParser
from .grpc.grpc_result_parser import GRPCResultParser
from .http.http_action_parser import HTTPActionParser
from .http.http_result_parser import HTTPResultParser
from .http2.http2_action_parser import HTTP2ActionParser
from .http2.http2_result_parser import HTTP2ResultParser
from .http3.http3_action_parser import HTTP3ActionParser
from .http3.http3_result_parser import HTTP3ResultParser
from .playwright.playwright_action_parser import PlaywrightActionParser
from .playwright.playwright_result_parser import PlaywrightResultParser
from .udp.udp_action_parser import UDPActionParser
from .udp.udp_result_parser import UDPResultParser
from .websocket.websocket_action_parser import WebsocketActionParser
from .websocket.websocket_result_parser import WebsocketResultParser



import json
from typing import Dict, Iterator, Union, List
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.action import HTTPAction

try:
    from graphql import Source, parse, print_ast

except ImportError:
    Source=None
    parse=lambda: None
    print_ast=lambda: None


class GraphQLAction(HTTPAction):

    def __init__(
        self,
        name: str, 
        url: str, 
        method: str = 'GET', 
        headers: Dict[str, str] = {}, 
        data: Union[str, dict, Iterator, bytes, None] = None, 
        user: str=None, 
        tags: List[Dict[str, str]] = [],
        redirects: int=3
    ) -> None:

        super(
            GraphQLAction,
            self
        ).__init__(
            name, 
            url, 
            method, 
            headers, 
            data, 
            user, 
            tags
        )

        self.type = RequestTypes.GRAPHQL
        self.redirects = redirects
        self.hooks: Hooks[GraphQLAction] = Hooks()

    def _setup_data(self) -> None:
        source = Source(self._data.get("query"))
        document_node = parse(source)
        query_string = print_ast(document_node)
        
        self.size = len(query_string)
        
        query = {
            "query": query_string
        }

        operation_name = self._data.get("operation_name")
        variables = self._data.get("variables")
        
        if operation_name:
            query["operationName"] = operation_name
        
        if variables:
            query["variables"] = variables

        self.encoded_data = json.dumps(query).encode()


from __future__ import annotations
import asyncio
import functools
import inspect
import multiprocessing as mp
import os
import random
import signal
import socket
import sys
from collections import defaultdict
from concurrent.futures import (
    ThreadPoolExecutor,
    ProcessPoolExecutor
)
from inspect import signature
from hedra.distributed.middleware.base import Middleware
from hedra.distributed.connection.tcp.mercury_sync_http_connection import MercurySyncHTTPConnection
from hedra.distributed.connection.tcp.mercury_sync_tcp_connection import MercurySyncTCPConnection
from hedra.distributed.connection.udp.mercury_sync_udp_connection import MercurySyncUDPConnection
from hedra.distributed.connection.udp.mercury_sync_udp_multicast_connection import MercurySyncUDPMulticastConnection
from hedra.distributed.env import load_env, Env
from hedra.distributed.models.base.error import Error
from hedra.distributed.models.base.message import Message
from pydantic import BaseModel
from typing import (
    Optional, 
    List, 
    Literal, 
    Union, 
    Dict,
    Any,
    Type,
    get_args,
    Callable,
    AsyncIterable,
    Tuple,
    TypeVarTuple,
    Generic
)
from types import MethodType
from .socket import (
    bind_tcp_socket,
    bind_udp_socket
)

P = TypeVarTuple('P')


mp.allow_connection_pickling()
spawn = mp.get_context("spawn")


def handle_worker_loop_stop(
    signame, 
    loop: asyncio.AbstractEventLoop,
    waiter: Optional[asyncio.Future]
):
    if waiter:
        waiter.set_result(None)

    loop.stop()


def handle_loop_stop(
    signame, 
    executor: Union[ProcessPoolExecutor, ThreadPoolExecutor],
):
        try:
            executor.shutdown(cancel_futures=True)

        except BrokenPipeError:
            pass

        except RuntimeError:
            pass


async def run(
    udp_connecton: MercurySyncUDPConnection,
    tcp_connection: MercurySyncTCPConnection,
    config: Dict[str, Union[int, socket.socket, str]]={}
):
    loop = asyncio.get_event_loop()
    
    waiter = loop.create_future()

    for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handle_worker_loop_stop(
                signame,
                loop,
                waiter
            )
        )  

    await udp_connecton.connect_async(
        cert_path=config.get('cert_path'),
        key_path=config.get('key_path'),
        worker_socket=config.get('udp_socket')
    )
    await tcp_connection.connect_async(
        cert_path=config.get('cert_path'),
        key_path=config.get('key_path'),
        worker_socket=config.get('tcp_socket')
    )


    await waiter


def start_pool(
    udp_connection: MercurySyncUDPConnection,
    tcp_connection: MercurySyncTCPConnection,
    config: Dict[str, Union[int, socket.socket, str]]={},
):
    import asyncio

    try:
        import uvloop
        uvloop.install()

    except ImportError:
        pass

    try:

        loop = asyncio.get_event_loop()

    except Exception:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    stdin_fileno = config.get('stdin_fileno')

    if stdin_fileno is not None:
        sys.stdin = os.fdopen(stdin_fileno)

    loop = asyncio.get_event_loop()
    
    loop.run_until_complete(
        run( 
            udp_connection,
            tcp_connection,
            config
        )
    )


class Controller(Generic[*P]):
    services: Dict[str, Type[Controller]] = {}

    def __init__(
        self,
        host: str,
        port: int,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        workers: int=0,
        env: Optional[Env]=None,
        engine: Literal["process", "async"]="async",
        middleware: List[Middleware]= []
    ) -> None:
        
        if env is None:
            env = load_env(Env)
    
        self.name = self.__class__.__name__
        self._instance_id = random.randint(0, 2**16)
        self._response_parsers: Dict[str, Message] = {}
        self._host_map: Dict[
            str, 
            Dict[
                Union[MercurySyncUDPConnection, MercurySyncTCPConnection],
                Tuple[str, int]
            ]
        ] = defaultdict(dict)

        if workers < 1:
            workers = 1

        self._workers = workers

        self.host = host
        self.port = port
        self.cert_path = cert_path
        self.key_path = key_path
        self.middleware = middleware

        self._env = env
        self._engine: Union[ProcessPoolExecutor, None] = None 
        self._udp_queue: Dict[Tuple[str, int], asyncio.Queue] = defaultdict(asyncio.Queue)
        self._tcp_queue: Dict[Tuple[str, int], asyncio.Queue] = defaultdict(asyncio.Queue)
        self._cleanup_task: Union[asyncio.Task, None] = None
        self._waiter: Union[asyncio.Future, None] = None

        self.engine_type = engine
        self._response_parsers: Dict[str, Message] = {}

        self.instance_ids = [
            self._instance_id + idx for idx in range(0, workers)
        ]

        if env.MERCURY_SYNC_USE_UDP_MULTICAST:
            self._udp = MercurySyncUDPMulticastConnection(
                self.host,
                self.port,
                self._instance_id,
                env=env
            ) 
        else:
            self._udp = MercurySyncUDPConnection(
                self.host,
                self.port,
                self._instance_id,
                env=env
            )
            
        if env.MERCURY_SYNC_USE_HTTP_SERVER:

            self._tcp = MercurySyncHTTPConnection(
                self.host,
                self.port + 1,
                self._instance_id,
                env=env
            ) 

        else:
            self._tcp =MercurySyncTCPConnection(
                self.host,
                self.port + 1,
                self._instance_id,
                env=env
            )

        self.setup()

    def setup(self):
        self.reserved_methods = [
            'connect',
            'send',
            'send_tcp',
            'stream',
            'stream_tcp',
            'close'
        ]

        middleware_enabled: Dict[str, bool] = {}

        response_parsers: Dict[
            str, 
            Callable[
                [Dict[str, Any]],
                BaseModel
            ]
        ] = {}
        controller_models: Dict[str, Message] = {}
        controller_methods: Dict[str, Callable[
            [Message],
            Message
        ]] = {}

        supported_http_handlers: Dict[str, Dict[str, str]] = defaultdict(dict)

        for _, method in inspect.getmembers(self, predicate=inspect.ismethod):
            (
                controller_models,
                controller_methods,
                middleware_enabled,
                response_parsers
            ) = self.apply_method(
                method,
                controller_models,
                controller_methods,
                middleware_enabled,
                response_parsers
            )
            

        self._parsers: Dict[str, Message] = {}
        self._events: Dict[str, Message] = {}

        for method_name, model in controller_models.items():

            self._udp.parsers[method_name] = model
            self._tcp.parsers[method_name] = model

            if isinstance(self._tcp, MercurySyncHTTPConnection):
                self._tcp._supported_handlers = supported_http_handlers
                self._tcp._middleware_enabled = middleware_enabled

            self._parsers[method_name] = model

        for method_name, method in controller_methods.items():

            self._udp.events[method_name] = method
            self._tcp.events[method_name] = method

            self._events[method_name] = method

        for key, parser in response_parsers.items():
            self._tcp._response_parsers[key] = parser

    def apply_method(
        self, 
        method: MethodType,
        controller_models: Dict[str, Message],
        controller_methods: Dict[str, Callable[
            [Message],
            Message
        ]],
        middleware_enabled: Dict[str, bool],
        response_parsers: Dict[
            str, 
            Callable[
                [Dict[str, Any]],
                BaseModel
            ]
        ]
    ) -> Tuple[
        Dict[str, Message],
        Dict[str, Callable[
            [Message],
            Message
        ]],
        Dict[str, bool],
        Dict[
            str, 
            Callable[
                [Dict[str, Any]],
                BaseModel
            ]
        ]
    ]:
        
        method_name = method.__name__

        not_internal = method_name.startswith('__') is False
        not_reserved = method_name not in self.reserved_methods
        is_server = hasattr(method, 'server_only')
        is_client = hasattr(method, 'client_only')
        is_http = hasattr(method, 'as_http') and method.as_http is True


        rpc_signature = signature(method)

        if not_internal and not_reserved and is_server:

            for param_type in rpc_signature.parameters.values():

                if issubclass(param_type.annotation, (BaseModel,)):

                    model = param_type.annotation
                    controller_models[method_name] = model

                controller_methods[method_name] = method

        elif not_internal and not_reserved and is_client:

            is_stream = inspect.isasyncgenfunction(method)

            if is_stream:

                response_type = rpc_signature.return_annotation
                args = get_args(response_type)

                response_call_type: Tuple[int, Message] = args[0]
                self._response_parsers[method.target] = get_args(response_call_type)[1]

            else:

                response_type = rpc_signature.return_annotation
                args = get_args(response_type)
                response_model: Tuple[int, Message] = args[1]

                self._response_parsers[method.target] = response_model

        if not_internal and not_reserved and is_http:

            path: str = method.path

            for middleware_operator in self.middleware:
                method = middleware_operator.wrap(method)
                middleware_enabled[path] = True

            response_type = rpc_signature.return_annotation
            args = get_args(response_type)

            response_model: Tuple[
                Union[BaseModel, str, None], 
                int
            ] = args[0]

            event_http_methods: List[str] = method.methods
            path: str = method.path

            for event_http_method in event_http_methods:
                event_key = f'{event_http_method}_{path}'

                for param_type in rpc_signature.parameters.values():
                    args = get_args(param_type.annotation)
                    
                    if len(args) > 0 and issubclass(args[0], (BaseModel,)):

                        path: str = method.path

                        model = args[0]

                        controller_models[event_key] = model

                controller_methods[event_key] = method

                if isinstance(method.responses, dict):

                    responses = method.responses

                    for status, status_response_model in responses.items():
                        status_key = f'{event_http_method}_{path}_{status}'

                        if issubclass(status_response_model, BaseModel):
                            response_parsers[status_key] = lambda response: status_response_model(
                                **response
                            ).json()

                if isinstance(method.serializers, dict):

                    serializers = method.serializers

                    for status, serializer in serializers.items():
                        status_key = f'{event_http_method}_{path}_{status}'

                        response_parsers[status_key] = serializer

        return (
            controller_models,
            controller_methods,
            middleware_enabled,
            response_parsers
        )
    
    async def run_forever(self):
        loop = asyncio.get_event_loop()
        self._waiter = loop.create_future()

        await self._waiter

    async def start_server(
      self,
      cert_path: Optional[str]=None,
      key_path: Optional[str]=None      
    ):
        
        for middleware in self.middleware:
            await middleware.__setup__()

        pool: List[asyncio.Future] = []

        loop = asyncio.get_event_loop()

        if self.engine_type == "process":
            engine = ProcessPoolExecutor(
                max_workers=self._workers,
                mp_context=mp.get_context(method='spawn')
            )

        if self.engine_type == 'process':

            udp_socket = bind_udp_socket(self.host, self.port)
            tcp_socket = bind_tcp_socket(self.host, self.port + 1)

            stdin_fileno: Optional[int]
            try:
                stdin_fileno = sys.stdin.fileno()
            except OSError:
                stdin_fileno = None

            config = {
                "udp_socket": udp_socket,
                "tcp_socket": tcp_socket,
                "stdin_fileno": stdin_fileno,
                "cert_path": cert_path,
                "key_path": key_path
            }


            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        engine
                    )
                )  

            for _ in range(self._workers):

                service_worker = loop.run_in_executor(
                    engine,
                    functools.partial(
                        start_pool,
                        MercurySyncUDPConnection(
                            self.host,
                            self.port,
                            self._instance_id,
                            self._env
                        ),
                        MercurySyncTCPConnection(
                            self.host,
                            self.port + 1,
                            self._instance_id,
                            self._env
                        ),
                        config=config
                    )
                )

                pool.append(service_worker) 
            
            await asyncio.gather(*pool)

        else:

            await self._udp.connect_async(
                cert_path=cert_path,
                key_path=key_path
            )

            await self._tcp.connect_async(
                cert_path=cert_path,
                key_path=key_path,
            )

    async def start_client(
        self,
        remotes: Dict[
            Tuple[str, int]: List[Type[Message]]
        ],
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None    
    ):

        for address, message_types in remotes.items():     

            host, port = address

            await self._tcp.connect_client(
                    (host, port + 1),
                    cert_path=cert_path,
                    key_path=key_path
            )

    async def send(
        self,
        event_name: str,
        message: Message
    ):
        shard_id, data = await self._udp.send(
            event_name,
            message.to_data(),
            (message.host, message.port)
        )

        if isinstance(data, Message):
            return shard_id, data

        response_data = self._response_parsers.get(event_name)(
            **data
        )

        return shard_id, response_data
    
    async def send_tcp(
        self,
        event_name: str,
        message: Message
    ):

        shard_id, data = await self._tcp.send(
            event_name,
            message.to_data(),
            (message.host, message.port + 1)
        )

        response_data = self._response_parsers.get(event_name)(
            **data
        )

        return shard_id, response_data
    
    async def stream(
        self,
        event_name: str,
        message: Message
    ) -> AsyncIterable[Tuple[int, Union[Message, Error]]]:
        
        address = (
            message.host,
            message.port
        )

        async for response in self._udp.stream(
            event_name,
            message.to_data(),
            address
        ):
            shard_id, data = response
            response_data = self._response_parsers.get(event_name)(
                **data
            )

            yield shard_id, response_data

    async def stream_tcp(
        self,
        event_name: str,
        message: Message
    ) -> AsyncIterable[Tuple[int, Union[Message, Error]]]:
        
        address = (
            message.host,
            message.port
        )

        async for response in self._tcp.stream(
            event_name,
            message.to_data(),
            address
        ):
            shard_id, data = response

            if data.get('error'):
                yield shard_id, Error(**data)

            response_data = self._response_parsers.get(event_name)(
                **data
            )

            yield shard_id, response_data

    async def close(self) -> None:

        if self._engine:
            self._engine.shutdown(cancel_futures=True)

        await self._udp.close()
        await self._tcp.close()

        if self._waiter:
            self._waiter.set_result(None)



import asyncio
import uuid
import warnings
from typing import (
    List, 
    Dict, 
    Any,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .mysql_connector_config import MySQLConnectorConfig


try:
    import sqlalchemy as sa

    # Aiomysql will raise warnings if a table exists despite us
    # explicitly passing "IF NOT EXISTS", so we're going to
    # ignore them.
    import aiomysql
    warnings.filterwarnings('ignore', category=aiomysql.Warning)

    from aiomysql.sa import create_engine, SAConnection
    from aiomysql.sa.result import RowProxy

    has_connector = True

except Exception:
    SAConnection = object
    RowProxy = object
    sqlalchemy = object
    sa = object
    create_engine = object
    CreateTable = object
    OperationalError = object
    has_connector = object



class MySQLConnector:
    connection_type=ConnectorType.MySQL

    def __init__(
        self, 
        config: MySQLConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.host = config.host
        self.database = config.database
        self.username = config.username
        self.password = config.password
        self.stage = stage
        self.parser_config = parser_config

        self.table_name =  config.table_name
        self._table = None

        self.metadata = sa.MetaData()
        self._engine = None
        self._connection = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        
        self.parser = Parser()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to MySQL instance at - {self.host} - Database: {self.database}')
        self._engine = await create_engine(
            db=self.database,
            host=self.host,
            user=self.username,
            password=self.password
        )

        self._connection: SAConnection = await self._engine.acquire()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to MySQL instance at - {self.host} - Database: {self.database}')
    
    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        if self._table is None:
            
            self._table = sa.Table(
                self.table_name,
                self.metadata,
                autoload_with=self._engine
            )
            
        rows: List[RowProxy] = [
            row async for row in self._connection.execute(
                self._table.select(**options)
            )
        ]

        return [
            {
              column: value  for column, value in result.items()
            } for result in rows
        ]
    
    async def close(self):
        await self._connection.close()

            